package example;

/*
 * DemoOde.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
import javax.robotics.engine.robots.DHRobot;
import javax.robotics.engine.robots.Robot;
import javax.robotics.ode.DerivableFunction;
import javax.robotics.ode.Euler;
import javax.robotics.ode.Heun;
import javax.robotics.ode.RungeKutta4th;
import javax.robotics.ode.Solver;
import javax.robotics.plot.MultiPlot;
import javax.robotics.plot.MultiPlotFrame;
import javax.robotics.swing.DisplayPanel;
import javax.robotics.vecmath.Matrix;
import javax.robotics.vecmath.RVector;
//import javax.swing.JFrame;

/**
 * Differential Equation integration demo.
 * @author Carmine Lia
 *
 */
public final class DemoOde implements DerivableFunction
{
    private final String newLine = System.getProperty("line.separator");
    
    private MultiPlotFrame frame;
    private MultiPlot plotPos, plotVel, plotPosRk4, plotVelRk4, plotPosH, plotVelH;
     
    private final int dof, npoints;
    private final double t0, tf;
    private RVector x; // state vector joints position + joints velocity
    private final RVector q0;

    private final Matrix K; // feedback gain matrix
    private final Robot robot;
    private final Solver euler, rk4, heun;
    private final DisplayPanel out = new DisplayPanel("Ode Demo Output");
    
    private DemoOde()
    {
        t0 = 0.0;
        tf = 4.0;
        npoints = 160;
        double dt = (tf - t0) / npoints;
        x = new RVector(4);

        double[] RR_data =
            { 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 2.0, -0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1666666, 0.0,
                    0.1666666, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0,
                    -0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0833333, 0.0, 0.0833333, 0.0, 0.0, 0.0, 0.0, 0.0 };

        robot = new DHRobot(2, RR_data);
        dof = robot.getDof();

        // Feedback gain matrix
        K = new Matrix(dof, 2 * dof);
        K.setElement(0, 0, 25.0);
        K.setElement(1, 1, 25.0);
        K.setElement(0, 2, 7.071);
        K.setElement(1, 3, 7.071);

        q0 = new RVector(2 * dof);
        q0.setElement(0, Math.PI / 4.0);
        q0.setElement(1, Math.PI / 4.0);

        heun = new Heun(dt, new RVector(x), this);
        euler = new Euler(dt, new RVector(x), this);
        rk4 = new RungeKutta4th(dt, new RVector(x), this);

        createPlotPanel();
        computeJointPos();
    }

    private void computeJointPos()
    {
        final Matrix J;
        final long startTime = System.nanoTime();

        out.append("=====================================================" + newLine
                        + "Robot RR dynamics" + newLine
                        + "=====================================================" + newLine);

        out.append("Robot joints variables" + newLine + robot.getQ());

        J = robot.inertia(robot.getQ());
        out.append(newLine + "Robot Inertia matrix" + newLine + J);
        out.append(newLine + "Feedback gain matrix" + newLine + K);

        out.append("(RungeKutta4th)" + newLine + " time\t  q1\t q2\t qp1\t qp2" + newLine);
     
        for (int i = 0; i <= npoints; i++)
        {

            x = euler.getLastValue();
            double time = euler.getTime();
            plotPos.addPoint(0, time, x.getElement(0), true);
            plotPos.addPoint(1, time, x.getElement(1), true);

            plotVel.addPoint(0, time, x.getElement(2), true);
            plotVel.addPoint(1, time, x.getElement(3), true);
            euler.step(x);

            x = heun.getLastValue();
            time = heun.getTime();

            plotPosH.addPoint(0, time, x.getElement(0), true);
            plotPosH.addPoint(1, time, x.getElement(1), true);

            plotVelH.addPoint(0, time, x.getElement(2), true);
            plotVelH.addPoint(1, time, x.getElement(3), true);
            heun.step(x);

            x = rk4.getLastValue();
            time = rk4.getTime();

            plotPosRk4.addPoint(0, time, x.getElement(0), true);
            plotPosRk4.addPoint(1, time, x.getElement(1), true);

            plotVelRk4.addPoint(0, time, x.getElement(2), true);
            plotVelRk4.addPoint(1, time, x.getElement(3), true);

            out.append(String.format("%6.3f\t", time));
            out.append(x.toString("%6.3f\t"));

            rk4.step(x);

        }
        plotPos.fillPlot();
        plotVel.fillPlot();
        plotPosH.fillPlot();
        plotVelH.fillPlot();
        plotPosRk4.fillPlot();
        plotVelRk4.fillPlot();

        final long estimatedTime = System.nanoTime() - startTime;
        out.append(newLine + "Estimated time:" + newLine + estimatedTime / 1e+9
                + " [s]");
    }
    // implementation of derivable function of Solver
    public RVector xDot(RVector xin)
    {
        final RVector q = new RVector(dof);
        final RVector qp = new RVector(dof);
        RVector qpp = new RVector(dof);
        final RVector tau = new RVector(dof);

        final RVector dx = new RVector(2 * dof);

        dx.sub(xin, q0);
        tau.mul(K, dx);

        tau.negate();

        xin.split(q, qp);
        qpp = robot.acceleration(q, qp, tau);

        final RVector xd = new RVector(qp);
        xd.append(qpp);
        return xd;
    }

    private void createPlotPanel()
    {
        frame = new MultiPlotFrame("ODE Demo", 3, 2);
   
        plotPos = new MultiPlot("Position (euler)", "[s]", "[rad]", "joint 1",
                "joint 2");
        plotPos.setXRange(0.0, tf);
        plotPos.setYRange(-1.0, 2);

        plotVel = new MultiPlot("Velocity (euler)", "[s]", "[rad/s]", "joint 1",
                "joint 2");
        plotVel.setXRange(0.0, tf);
        plotVel.setYRange(-1.0, 2);

        plotPosH = new MultiPlot("Position (heun)", "[s]", "[rad]", "joint 1",
                "joint 2");
        plotPosH.setXRange(0.0, tf);
        plotPosH.setYRange(-1.0, 2);
        plotVelH = new MultiPlot("Velocity (heun)", "[s]", "[rad/s]", "joint 1",
                "joint 2");
        plotVelH.setXRange(0.0, tf);
        plotVelH.setYRange(-1.0, 2.0);

        plotPosRk4 = new MultiPlot("Position (rk4)", "[s]", "[rad]", "joint 1",
                "joint 2");
        plotPosRk4.setXRange(0.0, tf);
        plotPosRk4.setYRange(-1.0, 2.0);
        plotVelRk4 = new MultiPlot("Velocity (rk4)", "[s]", "[rad/s]", "joint 1",
                "joint 2");
        plotVelRk4.setXRange(0.0, tf);
        plotVelRk4.setYRange(-1.0, 2.0);

        frame.add(plotPos);
        frame.add(plotVel);
        frame.add(plotPosH);
        frame.add(plotVelH);
        frame.add(plotPosRk4);
        frame.add(plotVelRk4);

        frame.setDefaultCloseOperation(MultiPlotFrame.EXIT_ON_CLOSE);
        frame.create();
    }

    public static void main(String[] args)
    {
        new DemoOde();
    }

}
