/*
 * TestControlRRA.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package example;

import java.io.IOException;
import java.net.URISyntaxException;

import javax.robotics.engine.controllers.EndEffController;
import javax.robotics.engine.controllers.ResolvedRateAcceleration;
import javax.robotics.engine.robots.DHRobot;
import javax.robotics.engine.robots.Robot;
import javax.robotics.engine.trajectory.Trajectory;
import javax.robotics.engine.trajectory.Trapezoidal;
import javax.robotics.ode.DerivableFunction2;
import javax.robotics.ode.RungeKutta4th2;
import javax.robotics.ode.Solver2;
import javax.robotics.plot.MultiPlot;
import javax.robotics.plot.MultiPlotFrame;
import javax.robotics.vecmath.RMatrix4d;
import javax.robotics.vecmath.RVector3d;
import javax.robotics.vecmath.RVector;

import org.apache.xmlbeans.XmlException;
/**
 * Solution of problem 6.16 of Modelling and Control of Robot Manipulators by
 * Sciavicco, Siciliano(2nd Edition) (compensated load mass)
 * 
 * @author Carmine Lia
 * 
 */
public class TestControlRRA implements DerivableFunction2
{
    private final ClassLoader cl = this.getClass().getClassLoader();
    
    MultiPlotFrame frame;
    MultiPlot plotErrPos, plotPos, plotEndEffPos, plotTorque;
    // numero dei giunti del robot
    int dof, npoints;
    double dur;
    RVector q, qp, tau, tf;
    RVector3d pd, ppd, pppd, pi, pf, delta, err;
    double s, sp, spp;
    Robot robot;
    Solver2 solver;
    Trajectory trj;
    EndEffController rra;

    public TestControlRRA() throws XmlException, IOException, URISyntaxException
    {

        dur = 2.50;
        double dt = 0.001;
        npoints = 2500;

        pd = new RVector3d();
        ppd = new RVector3d();
        pppd = new RVector3d();
        err = new RVector3d();
        delta = new RVector3d();

        qp = new RVector(2);
        tau = new RVector(2);

        // initial end eff position
        pi = new RVector3d(0.7, 0.2, 0.0);
        // final end eff position
        pf = new RVector3d(0.1, -0.6, 0.0);
        // max end eff velocity
        double vmax = 1.5;
        delta.sub(pf, pi);
        // init trajectory straight line from pi to pf
        trj = new Trapezoidal(0, 1, vmax / delta.length(), 1);

        // initial joints position
        q = new RVector(1.47653116289547, -2.39646300778072);
        // init robot setting the joint position at time 0
        robot = new DHRobot(cl.getResourceAsStream("example/conf/rr_book_chap6_load.xml"));
        dof = robot.getDof();
        robot.setQ(q);
        // System.out.println(robot.kine());

        // init solver with initial state
        solver = new RungeKutta4th2(dt, new RVector(q), new RVector(2), this);

        // set the gravity vector
        // System.out.println("Gravity vector:\n"+robot.getGravity());
        robot.setGravity(new RVector3d(0.0, 9.81, 0.0));
        System.out.println("Gravity vector:\n" + robot.getGravity());

        // init controller
        rra = new ResolvedRateAcceleration(robot, 16, 100);

        createPlotPanel();

        runControl();

    }
    private void createPlotPanel()
    {
        frame = new MultiPlotFrame("RRA control test", 2, 2);
     
        plotErrPos = new MultiPlot("End eff position error", "[s]", "[m]", "x", "y");
        plotErrPos.setXRange(0, dur);

        plotPos = new MultiPlot("Joints position", "[s]", "[rad]", "joint 1",
                "joint 2");
        plotPos.setXRange(0, dur);
        plotPos.setYRange(0, 1.6);

        plotEndEffPos = new MultiPlot("End Eff position", "[s]", "[m]", "x", "y");
        plotEndEffPos.setXRange(0, dur);

        plotTorque = new MultiPlot("Joints torque", "[s]", "[Nm]", "joint 1",
                "joint 2");
        plotTorque.setXRange(0, dur);

        frame.add(plotErrPos);
        frame.add(plotEndEffPos);
        frame.add(plotPos);
        frame.add(plotTorque);

        frame.setDefaultCloseOperation(MultiPlotFrame.EXIT_ON_CLOSE);
        frame.create();
        return;
    }

    private void runControl()
    {
        double time;
        RMatrix4d T = new RMatrix4d();
        final RVector3d p = new RVector3d();

        for (int i = 0; i < npoints; i++)
        {

            time = solver.getTime();
            q = solver.getLastValuePosition();
            qp = solver.getLastValueVelocity();
            // robot.setQ(q);
            T = robot.kine();
            p.x = T.m03;
            p.y = T.m13;

            // end eff path interpolation
            final  double[] path = trj.interpolate(time);
            pd.scaleAdd(path[0], delta, pi);
            ppd.scale(path[1], delta);
            pppd.scale(path[2], delta);

            tau = rra.torqueCmd(pppd, ppd, pd);
            err.sub(pd, p);

            plotErrPos.addPoint(0, time, err.x, true);
            plotErrPos.addPoint(1, time, err.y, true);

            plotEndEffPos.addPoint(0, time, p.x, true);
            plotEndEffPos.addPoint(1, time, p.y, true);

            plotPos.addVector(time, q, true);
            plotTorque.addVector(time, tau, true);

            solver.step(q, qp);

        }
        plotErrPos.fillPlot();
        plotPos.fillPlot();
        plotEndEffPos.fillPlot();
        plotTorque.fillPlot();
    }
    // implementation of derivable function of Solver
    public RVector xDot2(RVector q, RVector qp)
    {
        return robot.acceleration(q, qp, tau, new RVector3d(0, 0, 0),
                new RVector3d());
    }
    /**
     * @param args
     * @throws IOException 
     * @throws XmlException 
     * @throws URISyntaxException 
     */
    public static void main(String[] args) throws XmlException, IOException, URISyntaxException
    {
        new TestControlRRA();
    }

}
