/*
 * TestJointControlRobustSim.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package example;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.net.URISyntaxException;

import javax.robotics.engine.controllers.JointController;
import javax.robotics.engine.controllers.RobustControl;
import javax.robotics.engine.robots.DHRobot3D;
import javax.robotics.engine.robots.Robot3D;
import javax.robotics.engine.trajectory.Trajectory;
import javax.robotics.engine.trajectory.Trapezoidal;
import javax.robotics.ode.RungeKutta4th2;
import javax.robotics.ode.Solver2;
import javax.robotics.plot.MultiPlotPanel;
import javax.robotics.simulator.JointControlSimulation;
import javax.robotics.swing.DisplayPanel;
import javax.robotics.swing.RobotPanel;
import javax.robotics.vecmath.DMatrix;
import javax.robotics.vecmath.Matrix;
import javax.robotics.vecmath.RVector;
import javax.robotics.vecmath.RVector3d;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JPanel;

import org.apache.xmlbeans.XmlException;
/**
 * Solution of problem 6.13 of Modelling and Control of Robot Manipulators by
 * Sciavicco, Siciliano(2nd Edition) with Kp=diag(100,100), Kd=diag(25, 25)
 * 
 * @author Carmine Lia
 * 
 */
public class TestJointControlRobustSim extends JFrame implements ActionListener
{
    /**
     * 
     */
    private static final long serialVersionUID = -5758201787989857514L;
    
    // numero dei giunti del robot
    int dof, npoints;
    double dur;
    RVector q, qp, qd, qpd, qppd, tau, qs, qf, qpc, tf, err;
    Robot3D robot;
    Solver2 solver;
    Trajectory trj;
    JointController robust;
    
    private final JointControlSimulation sim;

    private final String newLine =  System.getProperty("line.separator");
    
    public TestJointControlRobustSim() throws XmlException, IOException, URISyntaxException
    {

        super("Robust control simulation");
        ClassLoader cl = this.getClass().getClassLoader();
        DisplayPanel display = new DisplayPanel("Simulation description");
        display.append("Robust Control"+newLine);
        display.append("Solution of problem 6.13 of Modelling and Control of Robot Manipulators by Sciavicco, Siciliano(2nd Edition) with Kp=diag(100,100), Kd=diag(25, 25)");
        dur = 1.50;
        double dt = 0.001;
        npoints = 1500;

        // initial joints position
        qs = new RVector(0.0, Math.PI / 4.0);
        q = new RVector(qs);

        // final joints position
        qf = new RVector(Math.PI / 2.0, Math.PI / 2.0);
        // max joints velocity
        qpc = new RVector(3.0 * Math.PI / 4.0, Math.PI / 3.0);

        // init trajectory
        trj = new Trapezoidal(qs, qf, qpc, 1.0);

        // init solver with initial state
        solver = new RungeKutta4th2(dt, new RVector(qs), new RVector(2));

        // init robot setting the joint position at time 0
        robot = new DHRobot3D(cl.getResourceAsStream("example/conf/rr_book_chap6_load.xml"));
       
   //     robot = new DHRobot3D("conf/rr_book_chap6_load.xml");
        dof = robot.getDof();

        // set the gravity vector
        // System.out.println("Gravity vector:\n"+robot.getGravity());
        robot.setGravity(new RVector3d(0.0, 9.81, 0.0));
      
        // init controller
        double[] inertia = {186.01, 0.0, 0.0,  132.5};
        robust = new RobustControl(robot, new DMatrix(100, 100), new DMatrix(25,
                25), new Matrix(2, inertia), 70.0, 0.001);

      
        sim = new JointControlSimulation(robot, robust, solver, trj, npoints);
        
        MultiPlotPanel plotsPanel = sim.getPlotPanel();
        plotsPanel.setLayout(new GridLayout(3,1));
        RobotPanel robotPanel = sim.getRobotPanel();
        
        final JPanel p = new JPanel();
        p.setLayout(new BorderLayout());
       
        final JButton b = new JButton("Play simulation");
        b.addActionListener(this);
         
        p.add(plotsPanel, BorderLayout.CENTER);
        p.add(b, BorderLayout.SOUTH);
        
        setPreferredSize(new Dimension(1024, 768));
        getContentPane().add(robotPanel, BorderLayout.CENTER);
        getContentPane().add(p, BorderLayout.EAST);
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        pack();
        
        setVisible(true);
        sim.setEnable(true);
    }

    
    
    public void actionPerformed(ActionEvent e)
    {
       sim.play();
    }


   
    
    /**
     * @param args
     * @throws IOException 
     * @throws XmlException 
     * @throws URISyntaxException 
     */
    public static void main(String[] args) throws XmlException, IOException, URISyntaxException
    {
        new TestJointControlRobustSim();
    }

}
