/*
 * ClikJacobianPosTranspose.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.engine.clik;

import javax.robotics.engine.robots.Robot;
import javax.robotics.vecmath.DMatrix;
import javax.robotics.vecmath.Matrix;
import javax.robotics.vecmath.RVector;
/**
 * Closed Loop Inverse Kinematics with positional Jacobian transpose. The class
 * implements the same scheme depicted in
 * {@link javax.robotics.engine.clik.ClikJacobianTranspose}: <br>
 * <p>
 * The orientation error is not computed.<br>
 * The integration method used is the forward Euler.
 * </p>
 * 
 * @author Carmine Lia
 * @since 1.0.2
 * @version 11/11/2005
 */
public class ClikJacobianPosTranspose extends ClikJacobianPosInverse
{
    private Matrix Jt;
    /**
     * Costructor with default sample time, dt = 1e-3.
     * 
     * @param robot
     *            the robot.
     * @param Kp
     *            the diagonal matrix of error position gain.
     *  @param spaceDim
     *            the dimension of robot operational space.          
     */
    public ClikJacobianPosTranspose(Robot robot, DMatrix Kp, int spaceDim)
    {
        super(robot, Kp, spaceDim);
        Jt = new Matrix(dof, spaceDim);
    }
    /**
     * Costructor.
     * 
     * @param robot
     *            the robot.
     * @param Kp
     *            the diagonal matrix of error position gain.
     * @param spaceDim
     *            the dimension of robot operational space.          
     * @param dt
     *            the sample time.
     */
    public ClikJacobianPosTranspose(Robot robot, DMatrix Kp, int spaceDim, double dt)
    {
        super(robot, Kp, spaceDim, dt);
        Jt = new Matrix(dof, spaceDim);
    }
    /**
     * Compute the joints position and velocity from a given desired end
     * effector position and velocity. The joints velocity is computed as:<br>
     * <img src="doc-files/eqtrasposta.png" alt = "q'=trans(J)*Ke"/><br>
     * the joints position is computed by Euler integration:<br>
     * q(t+dt) = q(t) + q'(t)*dt;
     * 
     * @param pd
     *            the vector of desired end effector position.
     * @param ppd
     *            the parameter is not considered.
     */
    public void step(RVector pd, RVector ppd)
    {
        // cinematica diretta
        robot.setQ(this.q);
        T = robot.kine();
        T.get(p_);
        p.set(p_);
        p.setSize(spaceDim);

        endEffPositionError(pd);

        xp.mul(Kp, posErr);
        // calcolo dello Jacobiano posizionale
        J = robot.jacobian();
        J.getSubMatrix(0, spaceDim - 1, 0, dof - 1, Jp);
        // trasposta
        Jt.transpose(Jp);
        this.qp.mul(Jt, xp);

        // trapezoidal integration
        /*
         * sum.add(this.qp, qp_prev); this.q.scaleAdd(h*0.5, sum, this.q);
         */
        // euler integration
        this.q.scaleAdd(h, this.qp, this.q);
    }
}
