/*
 * ClikJacobianTranspose.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.engine.clik;

import javax.robotics.engine.robots.Robot;
import javax.robotics.vecmath.*;
/**
 * Closed Loop Inverse Kinematics with Jacobian transpose. The class implements
 * this scheme:<br>
 * <img src=doc-files/trasposta.png alt="CLIK scheme with transpose" /> <br>
 * <p>
 * The orientation error is computed as quaternion error.<br>
 * The integration method used is the forward Euler.
 * </p>
 * 
 * @author Carmine Lia
 * @since 1.0.2
 * @version 11/11/2005
 */
public class ClikJacobianTranspose extends ClikJacobianInverse
{

    private Matrix Jt;
    /**
     * Costructor with default sample time, dt = 1e-3.
     * 
     * @param robot
     *            the robot.
     * @param Kp
     *            the diagonal matrix of error position gain.
     * @param Ko
     *            the diagonal matrix of error orientationo gain.
     */
    public ClikJacobianTranspose(Robot robot, DMatrix Kp, DMatrix Ko)
    {
        super(robot, Kp, Ko);
        Jt = new Matrix(dof, 6);
    }
    /**
     * Costructor.
     * 
     * @param robot
     *            the robot.
     * @param Kp
     *            the diagonal matrix of error position gain.
     * @param Ko
     *            the diagonal matrix of error orientationo gain.
     * @param dt
     *            the sample time.
     */
    public ClikJacobianTranspose(Robot robot, DMatrix Kp, DMatrix Ko, double dt)
    {
        super(robot, Kp, Ko, dt);
        Jt = new Matrix(dof, 6);
    }
    /**
     * Compute the joints position and velocity from a given desired end
     * effector position and orientation. The joints velocity is computed as:<br>
     * <img src="doc-files/eqtrasposta.png" alt = "q'=trans(J)*(K*e)"/><br>
     * the joints position is computed by Euler integration:<br>
     * q(t+dt) = q(t) + q'(t)*dt;
     * 
     * @param pd
     *            the vector of desired end effector position.
     * @param ppd
     *            input parameter not considered.
     * @param quatd
     *            the desired end effector quaternion.
     * @param wd
     *            input parameter not considered.
     */
    public void step(RVector3d pd, RVector3d ppd, Quaternion quatd, RVector3d wd)
    {
        // cinematica diretta
        robot.setQ(this.q);
        T = robot.kine();
        T.get(R, p);
        endEffPositionError(pd);
        endEffQuaternionError(quatd);

        xp.mul(Kp, posErr);
        xo.mul(Ko, orienErr);
        x = new RVector(xp);
        x.append(xo);

        J = robot.jacobian();
        Jt.transpose(J);
        this.qp.mul(Jt, x);
        // trapezoidal integration
        /*
         * sum.add(qp, qp_prev); q.scaleAdd(h*0.5, sum, q);
         */

        // euler integration
        this.q.scaleAdd(h, this.qp, this.q);

     }
}
