/*
 * JointStrokeConstraint.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.engine.clik;

import javax.robotics.vecmath.RVector;

/**
 * This class implements the constraint on joints stroke.<br>
 * <img src="doc-files/vincolovel.png" alt="q0'=k0*grad(w(q))" /> <br>
 * where<br>
 * <img src="doc-files/finecorsa.png" alt="w(q)" />
 * 
 * @author Carmine Lia
 * @since 1.0.2
 */
public class JointStrokeConstraint implements ConstraintVelocity
{
    private final RVector qavg, qdiff;
    private double k0;
    
    /**
     * Define the constraint with minimum and maximum position for each joint
     * and the constraint gain.
     * 
     * @param k0
     *            the constraint velocity gain.
     * @param qMax
     *            the vector of maximum positions.
     * @param qMin
     *            the vector of minimum positions.
     */
    public JointStrokeConstraint(double k0, RVector qMax, RVector qMin)
    {
        super();
        int size = qMax.getSize();
        this.k0 = k0;
        qavg = new RVector(size);
        qdiff = new RVector(size);
        qavg.add(qMax, qMin);
        qavg.scale(0.5);
        qdiff.sub(qMax, qMin);
    }

    final public double getK0()
    {
        return k0;
    }

    final public void setK0(double k0)
    {
        this.k0 = k0;
    }

    /**
     * Computes the constraint velocity function.<br>
     * <img src="doc-files/vincolovel.png" alt="q0'=k0*grad(w(q))" /> <br>
     * where<br>
     * <img src="doc-files/finecorsa.png" alt="w(q)" />.
     * 
     * @param q
     *            the joints position.
     * @return the current value of constraint.
     */
    public RVector compute(RVector q)
    {
        final int n = q.getSize();
        final RVector q0Dot = new RVector(n);
        for (int i = 0; i < n; i++)
        {
            final double num = (q.getElement(i) - qavg.getElement(i));
            final double den = (qdiff.getElement(i) * qdiff.getElement(i));
            final double val = num / den;
            final double val1 = 1.0 / n;
            q0Dot.setElement(i, -val1 * val * k0);
        }
        return q0Dot;
    }

}
