/*
 * ProportionalDerivative.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.engine.controllers;

import javax.robotics.engine.JRoboOp;
import javax.robotics.engine.robots.Robot;
import javax.robotics.vecmath.DMatrix;
import javax.robotics.vecmath.RVector;

/**
 * <p>
 * This class implements Proportional derivative with gravity compensation
 * joints position controller. The scheme of the control:<br>
 * <img src="doc-files/pdg.png" alt="scheme proportional derivative control with
 * gravity compensation" />
 * </p>
 * <p>
 * Wrapper class for C++ library ROBOOP.
 * </p>
 * 
 * @author <a href="http://www.intermedia.sa.it/lia/">Carmine Lia</a>
 * @since 1.0.0
 * @version 7/10/2005
 */
public class PDGravityCompensation extends JRoboOp implements JointController
{
    private short robotID;

    private short controllerID = 0;

    private DMatrix Kp, Kd;

    /**
     * Proportional Derivative with gravity compensation joints position
     * controller class constructor.
     * 
     * @param robot
     *            Robot class.
     * @param Kp
     *            diagonal matrix joint position error gain matrix.
     * @param Kd
     *            diagonal matrix joint velocity error gain matrix.
     */
    public PDGravityCompensation(Robot robot, double[] Kp, double[] Kd)
    {
        controllerID = JRoboOp.ID;
        robotID = robot.getID();

        this.Kp = new DMatrix(Kp);
        this.Kd = new DMatrix(Kd);
        createPDGravityComp(controllerID, robotID, Kp, Kd);
    }
    /**
     * Proportional Derivative with gravity compensation joints position
     * controller class constructor.
     * 
     * @param robot
     *            Robot class.
     * @param Kp
     *            diagonal matrix joint position error gain matrix.
     * @param Kd
     *            diagonal matrix joint velocity error gain matrix.
     */
    public PDGravityCompensation(Robot robot, DMatrix Kp, DMatrix Kd)
    {
        controllerID = JRoboOp.ID;
        robotID = robot.getID();
        double[] Kp_ = new double[Kp.getSize()];
        double[] Kd_ = new double[Kd.getSize()];
        Kp.get(Kp_);
        Kd.get(Kd_);
        this.Kp = Kp;
        this.Kd = Kd;
        createPDGravityComp(controllerID, robotID, Kp_, Kd_);
    }

    /**
     * Gets Controller class ID.
     * 
     * @return ID
     */
    public final short getID()
    {
        return controllerID;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.robotics.engine.Controller#getRobotID()
     */
    public final short getRobotID()
    {
        return robotID;
    }

    public final DMatrix getKp()
    {
        return this.Kp;
    }

    public final DMatrix getKd()
    {
        return this.Kd;
    }
    /**
     * Sets the position error gain matrix.
     * 
     * @param Kp
     *            position error gain diagonal matrix
     * @return 0 or WRONG_SIZE
     */
    public short setKp(double[] Kp)
    {
        this.Kp = new DMatrix(Kp);
        return setKpPDG(controllerID, robotID, Kp);

    }
    /**
     * Sets the velocity error gain matrix.
     * 
     * @param Kd
     *            velocity error gain diagonal matrix
     * @return 0 or WRONG_SIZE
     */
    public short setKd(double[] Kd)
    {
        this.Kd = new DMatrix(Kd);
        return setKdPDG(controllerID, robotID, Kd);

    }
    public short setKp(DMatrix Kp)
    {
        this.Kp = Kp;
        final double[] Kp_ = new double[Kp.getSize()];
        Kp.get(Kp_);
        return setKpPDG(controllerID, robotID, Kp_);
    }

    public short setKd(DMatrix Kd)
    {
        this.Kd = Kd;
        final double[] Kd_ = new double[Kd.getSize()];
        Kd.get(Kd_);
        return setKdPDG(controllerID, robotID, Kd_);
    }

    /**
     * Sets the output torque for a desired joint position.
     * 
     * @param qd
     *            array of desired joint position.
     * @return torque.
     */
    public double[] torqueCmd(double[]... qd)
    {
        return torqueCmdPDG(controllerID, robotID, qd[0]);
    }
    /**
     * Sets the output torque for a desired joint position and velocity.
     * 
     * @param qd
     *            the vector of desired joint position.
     * @return torque.
     */
    public RVector torqueCmd(RVector... qd)
    {

        final double[] qd_ = new double[qd[0].getSize()];
        qd[0].get(qd_);
        return new RVector(torqueCmdPDG(controllerID, robotID, qd_));

    }
}
