/*
 * ResolvedRateAcceleration.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.engine.controllers;

import javax.robotics.engine.JRoboOp;
import javax.robotics.engine.robots.Robot;
import javax.robotics.vecmath.Quaternion;
import javax.robotics.vecmath.RVector3d;
import javax.robotics.vecmath.RVector;

/**
 * <p>
 * This class implements Resolved Rate Acceleration controller. The
 * scheme of the control:<br>
 * <img src="doc-files/rra.png" alt="scheme of resolved rate acceleration
 * control" /> <br>
 * the dynamics model of robot:&nbsp;
 * {@code B(q)*q'' + n(q,q') = u,  n(q,q') = C(q,q')*q'+ Fq'+ g(q)} <br>
 * the resultant error equation:&nbsp; {@code e'' + Kd*e' + Kp*e = 0 }
 * </p>
 * Wrapper class for C++ library ROBOOP.
 * 
 * @author <a href="http://www.intermedia.sa.it/lia/">Carmine Lia</a>
 * @since 1.0.0
 * @version 7/10/2005
 * 
 */
public class ResolvedRateAcceleration extends JRoboOp implements
        EndEffController
{
    private short robotID;

    private short controllerID;

    private short dof;

    /**
     * Resolved Rate Acceleration controller class constructor.
     * 
     * @param robot
     *            Robot class.
     * @param Kdp
     *            end effector velocity error gain
     * @param Kpp
     *            end effector position error gain
     * @param Kdo
     *            end effector orientation angular rate gain
     * @param Kpo
     *            end effector orientation error gain
     */
    public ResolvedRateAcceleration(Robot robot, double Kdp, double Kpp,
            double Kdo, double Kpo)
    {
        controllerID = JRoboOp.ID;
        robotID = robot.getID();
        dof = (short) robot.getDof();
        createResolveRateAcceleration(controllerID, robotID, Kdp, Kpp, Kdo, Kpo);
    }
    /**
     * Resolved Rate Acceleration controller class constructor with positional
     * jacobian.
     * 
     * @param robot
     *            Robot class.
     * @param Kdp
     *            end effector velocity error gain
     * @param Kpp
     *            end effector position error gain
     */
    public ResolvedRateAcceleration(Robot robot, double Kdp, double Kpp)
    {
        controllerID = JRoboOp.ID;
        robotID = robot.getID();
        dof = (short) robot.getDof();
        createResolveRateAcceleration(controllerID, robotID, Kdp, Kpp, 0.0, 0.0);
    }
    /**
     * Gets Controller class ID.
     * 
     * @return ID
     */
    final public short getID()
    {
        return controllerID;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.robotics.engine.Controller#getRobotID()
     */
    final public short getRobotID()
    {
        return robotID;
    }
    
    public RVector3d getPositionError()
    {
         return new RVector3d(getPosErrorRRA(controllerID));
    }
    
    public RVector3d getOrientationError()
    {
        return new RVector3d(getQuatErrorRRA(controllerID));
    }
    
    /**
     * Sets the output torque for a desired end effector: acceleration,
     * velocity, position, angular acceleration, angular velocity, angular
     * position.
     * 
     * @param pdpp
     *            desired acceleration
     * @param pdp
     *            desired velocity
     * @param pd
     *            desired position
     * @param wdp
     *            desired angular acceleration
     * @param wd
     *            desired angular velocity
     * @param quatd
     *            desired vector angular position: quaternion (s,v1,v2,v3)
     * @return the output torque
     */
    public double[] torqueCmd(double[] pdpp, double[] pdp, double[] pd,
            double[] wdp, double[] wd, double[] quatd)
    {
        return torqueCmdRRA(controllerID, robotID, pdpp, pdp, pd, wdp, wd,
                quatd, dof);
    }
    /**
     * Sets the output torque for a desired end effector: acceleration,
     * velocity, position, angular acceleration, angular velocity, angular
     * position.
     * 
     * @param pdpp
     *            desired acceleration
     * @param pdp
     *            desired velocity
     * @param pd
     *            desired position
     * @param wdp
     *            desired angular acceleration
     * @param wd
     *            desired angular velocity
     * @param quatd
     *            desired vector angular position: quaternion (s,v1,v2,v3)
     * @return the output torque
     */
    public RVector torqueCmd(RVector3d pdpp, RVector3d pdp, RVector3d pd,
            RVector3d wdp, RVector3d wd, Quaternion quatd)
    {
        final double[] pdpp_ = new double[3];
        final double[] pdp_ = new double[3];
        final double[] pd_ = new double[3];
        final double[] wdp_ = new double[3];
        final double[] wd_ = new double[3];
        final double[] quatd_ = new double[4];

        pdpp.get(pdpp_);
        pdp.get(pdp_);
        pd.get(pd_);
        wdp.get(wdp_);
        wd.get(wd_);
        quatd_[0] = quatd.w;
        quatd_[1] = quatd.x;
        quatd_[2] = quatd.y;
        quatd_[3] = quatd.z;

        return new RVector(torqueCmdRRA(controllerID, robotID, pdpp_, pdp_, pd_,
                wdp_, wd_, quatd_, dof));
    }

    /**
     * Sets the output torque for a desired end effector: acceleration,
     * velocity, position. Only the positional jacobian is used.
     * 
     * @param pdpp
     *            desired acceleration
     * @param pdp
     *            desired velocity
     * @param pd
     *            desired position
     * @return the output torque
     */
    public RVector torqueCmd(RVector3d pdpp, RVector3d pdp, RVector3d pd)
    {
        final double[] pdpp_ = new double[3];
        final double[] pdp_ = new double[3];
        final double[] pd_ = new double[3];

        pdpp.get(pdpp_);
        pdp.get(pdp_);
        pd.get(pd_);

        return new RVector(torqueCmdRRA(controllerID, robotID, pdpp_, pdp_, pd_));
    }

}