package javax.robotics.j3d;

/*
 *  Cylinder - Shape3D Cylinder
 *  Copyright (C) 24 aprile 2003 Carmine Lia
 *  email: carmine.lia@libero.it
 *  webaddress: http://digilander.libero.it/carmine.lia/
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
import java.util.*;
import com.sun.j3d.utils.geometry.*;
import javax.media.j3d.*;
import javax.vecmath.*;

/**
 * Cilindro con la possibilit? di allineare la generatrice con uno degli assi
 * coordinati.
 * 
 * @author Carmine Lia
 * @version 26 aprile 2003
 */
public class Cylinder extends Shape3D
{
   // private final static boolean DEBUG = false;

    // opzioni
    // sviluppo lungo l'asse ...
    /**
     * Direzione X della generatrice del cilindro
     */
    public final static int X = 1;
    /**
     * Direzione Y della generatrice del cilindro
     */
    public final static int Y = 2;
    /**
     * Direzione Z della generatrice del cilindro
     */
    public final static int Z = 3;

    // disegna solo il ...
    /**
     * Disegno solo della superficie del cilindro
     */
    public final static int BODY = 4;
    // disegna con il ...
    /**
     * Disegno con la parte superiore chiusa
     */
    public final static int TOP = 5;
    /**
     * Disegno con la parte inferiore chiusa
     */
    public final static int BOTTOM = 6;
    /**
     * Disegno con parte inferiore e parte superiore chiuse
     */
    public final static int TOP_BOTTOM = 7;

    /**
     * Costruttore di un cilindro di raggio 1.0 e lunghezza 2.0 che si sviluppa
     * lungo l'asse Z.
     */
    public Cylinder()
    {
        // raggio di default
        double r1 = 1.0;
        double r2 = r1;
        // lunghezza di default
        double l = 2.0;
        // num di divisioni
        int div = 20;
        // posizione del centro della base cilindro
        Point3d c = new Point3d(0.0, 0.0, -l / 2.0);

        this.setGeometry(createGeometry(c, r1, r2, l, div, BODY, Z));
    }

    /**
     * Costruttore di un cilindro di raggio r e lunghezza l che si sviluppa
     * lungo l'asse Z.
     * 
     * @param c
     *            Coordinata del centro della base
     * @param r
     *            Raggio del cilindro
     * @param l
     *            Lunghezza del cilindro
     */
    public Cylinder(Point3d c, double r, double l)
    {
        // num di divisioni
        int div = 20;
        double r2 = r;
        this.setGeometry(createGeometry(c, r, r2, l, div, BODY, Z));

    }

    /**
     * Costruttore di un cilindro di raggio r e lunghezza l che si sviluppa
     * lungo l'asse Z.
     * 
     * 
     * @param c
     *            Coordinata del centro della base
     * @param r
     *            Raggio del cilindro
     * @param l
     *            Lunghezza del cilindro.
     * @param prim
     *            Tipo di cilindro: BODY, TOP, BOTTOM, TOP_BOTTOM.
     */
    public Cylinder(Point3d c, double r, double l, int prim)
    {
        // num di divisioni
        int div = 20;
        double rt = r;
        this.setGeometry(createGeometry(c, r, rt, l, div, prim, Z));

    }

    /**
     * Costruttore di un cilindro di raggio r e lunghezza l.
     * 
     * @param c
     *            Coordinata del centro della base
     * @param r
     *            Raggio della circonferenza della base
     * @param l
     *            Lunghezza del cilindro.
     * @param prim
     *            Tipo di cilindro: BODY, TOP, BOTTOM, TOP_BOTTOM.
     * @param gen
     *            Direzione della generatrice: X, Y, Z.
     */
    public Cylinder(Point3d c, double r, double l, int prim, int gen)
    {
        // num di divisioni
        int div = 20;
        double r2 = r;
        this.setGeometry(createGeometry(c, r, r2, l, div, prim, gen));

    }

    /**
     * Costruttore di un cilindro di raggio r e lunghezza l.
     * 
     * @param c
     *            Coordinata del centro della base
     * @param r
     *            Raggio del cilindro.
     * @param l
     *            Lunghezza del cilindro.
     * @param prim
     *            Tipo di cilindro: BODY, TOP, BOTTOM, TOP_BOTTOM.
     * @param gen
     *            Direzione della generatrice: X, Y, Z.
     * @param div
     *            Numero di divisioni.
     */
    public Cylinder(Point3d c, double r, double l, int gen, int prim, int div)
    {
        double r2 = r;
        this.setGeometry(createGeometry(c, r, r2, l, div, prim, gen));

    }

    // ///////////////////////////////////////////////////////////////
    //
    // costruttori con raggio di base e raggio della parte superiore
    //

    /**
     * Costruttore di un cilindro di raggi rb, rt e lunghezza l che si sviluppa
     * lungo l'asse Z.
     * 
     * @param c
     *            Coordinata del centro della base
     * @param rb
     *            Raggio della circonferenza della base
     * @param rt
     *            Raggio della circonferenza della parte superiore
     * @param l
     *            Lunghezza del cilindro
     */
    public Cylinder(Point3d c, double rb, double rt, double l)
    {
        // num di divisioni
        int div = 20;
        this.setGeometry(createGeometry(c, rb, rt, l, div, BODY, Z));

    }

    /**
     * Costruttore di un cilindro di raggi rb, rt e lunghezza l che si sviluppa
     * lungo l'asse Z.
     * 
     * @param c
     *            Coordinata del centro della base
     * @param rb
     *            Raggio della circonferenza della base
     * @param rt
     *            Raggio della circonferenza della parte superiore
     * @param l
     *            Lunghezza del cilindro.
     * @param prim
     *            Tipo di cilindro: BODY, TOP, BOTTOM, TOP_BOTTOM.
     */
    public Cylinder(Point3d c, double rb, double rt, double l, int prim)
    {
        // num di divisioni
        int div = 20;
        this.setGeometry(createGeometry(c, rb, rt, l, div, prim, Z));

    }

    /**
     * Costruttore di un cilindro di raggi rb, rt e lunghezza l.
     * 
     * @param c
     *            Coordinata del centro della base
     * @param rb
     *            Raggio della circonferenza della base
     * @param rt
     *            Raggio della circonferenza della parte superiore
     * @param l
     *            Lunghezza del cilindro.
     * @param prim
     *            Tipo di cilindro: BODY, TOP, BOTTOM, TOP_BOTTOM.
     * @param gen
     *            Direzione della generatrice: X, Y, Z.
     */
    public Cylinder(Point3d c, double rb, double rt, double l, int prim, int gen)
    {
        // num di divisioni
        int div = 20;
        this.setGeometry(createGeometry(c, rb, rt, l, div, prim, gen));

    }

    /**
     * Costruttore di un cilindro di raggi rb, rt e lunghezza l.
     * 
     * @param c
     *            Coordinata del centro della base
     * @param rb
     *            Raggio della circonferenza di base.
     * @param rt
     *            Raggio della circonferenza della parte superiore.
     * @param l
     *            Lunghezza del cilindro.
     * @param prim
     *            Tipo di cilindro: BODY, TOP, BOTTOM, TOP_BOTTOM.
     * @param gen
     *            Direzione della generatrice: X, Y, Z.
     * @param div
     *            Numero di divisioni.
     */
    public Cylinder(Point3d c, double rb, double rt, double l, int gen,
            int prim, int div)
    {
        this.setGeometry(createGeometry(c, rb, rt, l, div, prim, gen));
    }
    // ////////////////////////////////////////////////////////////////////////////////////////////////////
    // fine elenco costruttori

    /**
     * Creazione del cilindro
     * 
     * @param c
     *            Coordinate del centro della base del cilindro
     * @param l
     *            Lunghezza del cilindro
     * @param div
     *            Numero di divisioni
     * @param prim
     *            Tipo di cilindro: BODY, TOP, BOTTOM, TOP_BOTTOM(default)
     * @param gen
     *            Direzione della generatrice: X, Y, Z
     * @param r1
     *            Raggio della base
     * @param r2
     *            Raggio della parte superiore
     * @return
     */
    private Geometry createGeometry(Point3d c, double r, double r2,
            double l, int div, int prim, int gen)
    {
        final double dteta = (2.0 * Math.PI) / div;
        // punti della base TODO usare ArrayList o FastArray
        final Point3d[] base = new Point3d[div];
        // punti della copertura
        final Point3d[] top = new Point3d[div];
        Point3d[] data = null;
        int[] stripCount = null;
        // inizializzazione dei vettori dei dati
        switch (prim)
        {
        case BODY:
            data = new Point3d[4 * div];
            stripCount = new int[div];
            break;
        case TOP:
            data = new Point3d[4 * div + div];
            stripCount = new int[div + 1];
            break;
        case BOTTOM:
            data = new Point3d[4 * div + div];
            stripCount = new int[div + 1];
            break;
        default:
            data = new Point3d[4 * div + 2 * div];
            stripCount = new int[div + 2];
            break;

        }
         // punti dischi di base e superiore
        for (int i = 0; i < div; i++)
        {
            final double teta = i * dteta;
            final double sinTeta = Math.sin(teta);
            final double cosTeta = Math.cos(teta);
            
            switch (gen)
            {
            case X:
                base[i] = new Point3d(c.x, c.y + r * sinTeta, c.z+r * cosTeta);
                top[i] = new Point3d(l + c.x, c.y+r2 * sinTeta, c.z+r2 * cosTeta);
                break;
            case Y:
                base[i] = new Point3d(c.x+r * cosTeta, c.y, c.z+r * sinTeta);
                top[i] =  new Point3d(c.x+r2 * cosTeta, l + c.y, c.z+r2 * sinTeta);
                break;
            case Z:
                base[i] = new Point3d(c.x+r * sinTeta, c.y+r * cosTeta, c.z);
                top[i] = new Point3d(c.x+r2 * sinTeta, c.y +r2 * cosTeta, l + c.z);
                break;
            }

        }

        // creazione della superficie del cilindro
        int j = 0;
        for (int k = 0; k < div; k++)
        {
            data[j] = new Point3d(base[k]);
            data[++j] = new Point3d(top[k]);
            if(k < (div - 1))
            {
                data[++j] = new Point3d(top[k + 1]);
                data[++j] = new Point3d(base[k + 1]);
            }
            else
            {
                data[++j] = new Point3d(top[0]);
                data[++j] = new Point3d(base[0]);
            }
            j++;

        }
        switch (prim)
        {
        case BODY:
            Arrays.fill(stripCount, 4);
            break;
        case BOTTOM:
            // creazione del disco di base
            for (int k = 0; k < div; k++)
            {
                data[j++] = new Point3d(base[k]);
            }
            Arrays.fill(stripCount, 4);
            stripCount[div] = div;
            break;
        case TOP:
            // creazione del disco superiore
            for (int k = div - 1; k >= 0; k--)
            {
                // for(int k=0; k<div; k++)
                data[j++] = new Point3d(top[k]);
            }
            Arrays.fill(stripCount, 4);
            stripCount[div] = div;
            break;
        default:
            // creazione del disco di base
            for (int k = 0; k < div; k++)
            {
                data[j++] = new Point3d(base[k]);
            }
            // disco superiore
            for (int k = div - 1; k >= 0; k--)
            {
                data[j++] = new Point3d(top[k]);
            }
            Arrays.fill(stripCount, 4);
            stripCount[div] = div;
            stripCount[div + 1] = div;
            break;
        }
        /*
         * if(DEBUG) { for (int i = 0; i < data.length; i++) {
         * System.out.println("i: " + i + "\n\tx: " + data[i].x + "\n\ty: " +
         * data[i].y + "\n\tz: " + data[i].z + "\n"); } }
         */
        final GeometryInfo gi = new GeometryInfo(GeometryInfo.POLYGON_ARRAY);
        gi.setCoordinates(data);
        gi.setStripCounts(stripCount);
        if(l < 0)
            gi.reverse();

        final NormalGenerator ng = new NormalGenerator();
        ng.generateNormals(gi);
        gi.recomputeIndices();

        final Stripifier st = new Stripifier();
        st.stripify(gi);
        gi.recomputeIndices();

        return gi.getGeometryArray();
    }
    // fine del metodo createGeometry

}
