/*
 * Euler.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.ode;

import javax.robotics.vecmath.RVector;




/**
 * <p>
 * This class implements forward Euler integration method.  
 * </p>
 * 
 * @author <a href="http://www.intermedia.sa.it/lia/">Carmine Lia</a>
 * @since 1.0.0
 * @version 09/11/2005
 */
public class Euler implements Solver, DerivableFunction
{
    RVector lastValue;
    double h, time;
    
    private final DerivableFunction target;
    private RVector initialCondition;
    
    /**
     * Defines numerical integration for x' = f(x) differential equation.
     * @param dt time interval of integration [sec]
     * @param initialCondition initial condition at time zero x(0).
     * @param target differential equation to be integrated
     */
    public Euler(double dt, RVector initialCondition, DerivableFunction target)
    {
        this.h = dt;
        lastValue = new RVector(initialCondition);
        this.initialCondition = initialCondition;
        this.target = target;
        time = startTime;
    }
    
    /* (non-Javadoc)
     * @see javax.robotics.engine.Solver#setSampleTime(double)
     */
    public final void setSampleTime(double dt)
    {
        this.h = dt;
    }
    
    public final double getSampleTime()
    {
       return this.h;
    }
    
    /* (non-Javadoc)
     * @see javax.robotics.engine.Solver#setInitialCondition(RVector)
     */
    public final void setInitialCondition(RVector initialCondition)
    {
        lastValue = new RVector(initialCondition);
        this.initialCondition = initialCondition;
    }
    
    /* (non-Javadoc)
     * @see javax.robotics.engine.Solver#getInitialCondition()
     */
    public final RVector getInitialCondition()
    {
        return initialCondition;
    }
   
    /* (non-Javadoc)
     * @see javax.robotics.engine.Solver#getLastValue()
     */
    public final RVector getLastValue()
    {
        return lastValue;
    }
   
    /* (non-Javadoc)
     * @see javax.robotics.engine.Solver#getTime()
     */
    public final double getTime()
    {
        return time;
    }
   
    /* (non-Javadoc)
     * @see javax.robotics.engine.Solver#step(RVector)
     */
    public void step(RVector x)
    {
     // x' = f(x)
      final RVector xdot = xDot(x);
      
      // x(t+dt) = x(t) + dt*f(x)
      lastValue.scaleAdd(h, xdot, lastValue);
 
      time += h;
    }
   
    /* (non-Javadoc)
     * @see javax.robotics.engine.DerivableFunction#xDot(double[])
     */
    public RVector xDot(RVector x)
    {
        final RVector temp;
        temp = (target != null)? target.xDot(x):null;
        return temp;
    }
    
}
