/*
 * RungeKutta4th2.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.ode;

import javax.robotics.vecmath.RVector;
import javax.vecmath.MismatchedSizeException;
/**
 * <p>
 * This class implements forward RungeKutta4th integration method for systems of
 * two first order differential equations.
 * <p>
 * <code>x''=f(x, x') with x(0)=x<sub>0</sub>, x'(0)= y<sub>0</sub>, is equivalent to system of two first order of differential equation:</br>
 * y' = f(x, y)</br>
 * x' = y</br>
 * with x(0)=x<sub>0</sub>, y(0)= y<sub>0</sub>
 * </code>
 * </p>
 * </p>
 * 
 * @author <a href="http://www.intermedia.sa.it/lia/">Carmine Lia</a>
 * @since 1.0.3
 * @version 09/11/2005
 */
public class RungeKutta4th2 extends Euler2 
{
    private RVector k1, k2, k3, k4, l1, l2, l3, l4, x1, x2, x3, y1, y2, y3;
    private int size;

    /**
     * Defines RungeKutta4th numerical integration for x'' = f(x, x') as system
     * of two first order differential equation.
     * 
     * @param dt
     *            time interval of integration [sec]
     * @param initialConditionPosition
     *            position at time zero x(0).
     * @param initialConditionVelocity
     *            velocity at time zero x'(0).
     */
    public RungeKutta4th2(double dt, RVector initialConditionPosition,
            RVector initialConditionVelocity)
    {
        this(dt, initialConditionPosition, initialConditionVelocity, null);
    }
    
    /**
     * Defines RungeKutta4th numerical integration for x'' = f(x, x') as system
     * of two first order differential equation.
     * 
     * @param dt
     *            time interval of integration [sec]
     * @param initialConditionPosition
     *            position at time zero x(0).
     * @param initialConditionVelocity
     *            velocity at time zero x'(0).
     * @param target
     *            differential equation to be integrated
     */
    public RungeKutta4th2(double dt, RVector initialConditionPosition,
            RVector initialConditionVelocity, DerivableFunction2 target)
    {
        super(dt, initialConditionPosition, initialConditionVelocity, target);

        size = initialConditionPosition.getSize();

        if(size != initialConditionVelocity.getSize())
            throw new MismatchedSizeException(
                    "The size of the  position and velocity vectors must be equals");

        k1 = new RVector(size);
        k2 = new RVector(size);
        k3 = new RVector(size);
        k4 = new RVector(size);

        l1 = new RVector(size);
        l2 = new RVector(size);
        l3 = new RVector(size);
        l4 = new RVector(size);

        x1 = new RVector(size);
        x2 = new RVector(size);
        x3 = new RVector(size);

        y1 = new RVector(size);
        y2 = new RVector(size);
        y3 = new RVector(size);

    }

    public void step(RVector position, RVector velocity)
    {
        final double sei = 1.0/6.0;
        final double tre = 1.0/3.0;
        final double[] scale = {1.0, sei, tre, tre, sei };
        
        // x'' = f(x, x')
        RVector xdot = xDot2(position, velocity);
        
        // k1 = h*x'
        k1.scale(h, velocity);
        x1.scaleAdd(0.5, k1, position);
        
        // l1 = h*f(x, x')
        l1.scale(h, xdot);
        y1.scaleAdd(0.5, l1, velocity);
        
        // k2 = h*(x' + 0.5*l1)
        k2.scale(h, y1);
        x2.scaleAdd(0.5, k2, position);
        
        // l2 = h*f(x+0.5*k1, x'+ 0.5*l1)
        xdot = xDot2(x1, y1);
        l2.scale(h, xdot);
        y2.scaleAdd(0.5, l2, velocity);
        
        //  k3 = h*(x' + 0.5*l2)
        k3.scale(h, y2);
        x3.add(k3, position);
        
        // l3 = h*f(x+0.5*k2, x'+ 0.5*l2)
        xdot = xDot2(x2, y2);
        l3.scale(h, xdot);
        y3.add(l3, velocity);
        
        //  k4 = h*(x' + l3)
        k4.scale(h, y3);
        
        // l4 = h*f(x+ k3, x'+ l3)
        xdot = xDot2(x3, y3);
        l4.scale(h, xdot);
        
               
        lastValuePosition.addScale(scale, lastValuePosition, k1, k2, k3, k4);
        lastValueVelocity.addScale(scale, lastValueVelocity, l1, l2, l3, l4);
        
        time += h;

    }

}
