/*
 * MultiPlotPanel.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.plot;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridLayout;
import java.awt.geom.AffineTransform;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;

import javax.swing.JPanel;

import javolution.util.FastTable;

/**
 * Class to create multiple plot panel. Example.
 * 
 * <pre>
 *    {@code
 *    
 *    ...
 *    MultiPlotPanel mp = new MultiPlotPanel();
 *    MultiPlot mp1 = new MultiPlot();
 *    MultiPlot mp2 = new MultiPlot();
 *    MultiPlot mp3 = new MultiPlot();
 *    MultiPlot mp4 = new MultiPlot();
 *    mp.add(mp1);
 *    mp.add(mp2);
 *    mp.add(mp3);
 *    mp.add(mp4);
 *    mp.create();
 *    ...
 *    
 *    }
 * </pre>
 * 
 * @author <a href="http://www.intermedia.sa.it/lia/">Carmine Lia</a>
 * @since 1.1.2
 * @version 01/11/2005
 */
public class MultiPlotPanel extends JPanel implements Printable
{
    /**
     * 
     */
    private static final long serialVersionUID = 5275119495930756497L;
    private static final int maxPlots = 6;
    private final String title;
    private final FastTable<MultiPlot> plots = new FastTable<MultiPlot>(
            maxPlots);
    /**
     * Default constructor of a 2x2 grid of plots.
     * 
     */
    public MultiPlotPanel()
    {
        this("", 2, 2);
    }
    
    /**
     * Default constructor of a 2x2 grid of plots.
     * 
     * @param title
     *            the description to be printed.
     */
    public MultiPlotPanel(String title)
    {
        this(title, 2, 2);
    }

    /**
     * Constructor of a plots grid with the specified number of rows and
     * columns.
     * 
     * @param title
     *            the description to be printed
     * @param nRows
     *            the number of rows.
     * @param nCols
     *            the number of columns.
     */
    public MultiPlotPanel(String title, int nRows, int nCols)
    {
        super();
        // this.setPreferredSize(new Dimension(800, 600));
        this.title = title;
        this.setBackground(Color.WHITE);
        setBackground(Color.WHITE);
        setLayout(new GridLayout(nRows, nCols));
    }
    /**
     * Adds a new plot to the panel.
     * 
     * @param mp
     *            the plot.
     */
    public void add(MultiPlot mp)
    {
        super.add(mp);
        plots.add(mp);
    }
    public void showButtons(boolean show)
    {
        final int size = plots.size();
        for (int i = 0; i < size; i++)
        {
            plots.get(i).setButtons(show);
        }

    }

    public void resetPlots()
    {
        final int size = plots.size();
        for (int i = 0; i < size; i++)
        {
            plots.get(i).resetAxes();
        }

    }

    public void fillPlots()
    {
        final int size = plots.size();
        for (int i = 0; i < size; i++)
        {
            plots.get(i).fillPlot();
        }

    }
    public int print(Graphics graphics, PageFormat pageFormat, int pageIndex)
            throws PrinterException
    {

        if(graphics == null)
        {
            return Printable.NO_SUCH_PAGE;
        }

        // We only print on one page.
        if(pageIndex >= 1)
        {
            return Printable.NO_SUCH_PAGE;
        }
        final Graphics2D graphics2D = (Graphics2D) graphics;

        final AffineTransform t2d = new AffineTransform();
        // Scale the printout to fit the pages.
        final double scalex = pageFormat.getImageableWidth()
                / (double) this.getWidth();
        final double scaley = pageFormat.getImageableHeight()
                / (double) this.getHeight();
        final double scale = Math.min(scalex, scaley);
        t2d.translate((int) pageFormat.getImageableX(), (int) pageFormat
                .getImageableY());
        t2d.scale(scale, scale);
        graphics2D.transform(t2d);
        this.paint(graphics2D);

        graphics2D.drawString(title, 1.0f, (float) pageFormat
                .getImageableHeight() - 1.0f);
        return Printable.PAGE_EXISTS;

    }
}
