/*
 * CLIKSimulation.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.simulator;

import java.util.Enumeration;
import java.util.Timer;
import java.util.TimerTask;

import javax.media.j3d.Behavior;
import javax.media.j3d.WakeupCriterion;
import javax.media.j3d.WakeupOnElapsedTime;
import javax.robotics.engine.clik.CLIK;
import javax.robotics.engine.robots.Robot3D;
import javax.robotics.engine.trajectory.Trajectory;
import javax.robotics.plot.MultiPlot;
import javax.robotics.plot.MultiPlotPanel;
import javax.robotics.swing.RobotPanel;
import javax.robotics.vecmath.Quaternion;
import javax.robotics.vecmath.RVector3d;
import javax.robotics.vecmath.RVector;

/**
 * @author carmine lia
 * @version 01/nov/06
 *
 */
public class CLIKSimulation extends Behavior
{
    Robot3D robot;
    private CLIK clik;
    private Trajectory trj;
    private WakeupCriterion wakeUp;

    // simulation data
    private int points;
    private double time = 0.0;
    private double timeStep;
    private final Timer sceneTimer;
  

    // output data
    int i = 0;
    double jointData[][];

    // robot data
    private int dof;
    private final RVector q, qp;

    // trajectory data
    private final RVector3d s = new RVector3d();
    private final RVector3d ds = new RVector3d();
    private final RVector3d dds = new RVector3d();
    private final Quaternion quatd = new Quaternion();
    private final RVector3d wd = new RVector3d();

    // plot data
    private final MultiPlotPanel plotPanel;
    private final MultiPlot plotErrOrien;
    private final MultiPlot plotJVel;
    private final MultiPlot plotEndEffErrPos;
    private final MultiPlot plotJPos;

    // robot scene 3d
    private final RobotPanel robotPanel;

    private static final long updateSchedulerTime = 150l;
    static final int deltaTimer = 20;
    private double timeDuration;

    public CLIKSimulation(Robot3D robot, CLIK clik, Trajectory trj, int points)
    {
        this.robot = robot;
        this.clik = clik;
        this.trj = trj;
        this.points = points;
        this.timeStep = clik.getSampleTime();
        timeDuration = points * timeStep;
        dof = robot.getDof();
        jointData = new double[points][dof];

        // initialize joints
        q = clik.getJointsPosition();
        qp = clik.getJointsVelocity();

        sceneTimer = new Timer();
        wakeUp = new WakeupOnElapsedTime((int) Math
                .round(clik.getSampleTime() * 1000.0));

        // plots panel
        plotPanel = new MultiPlotPanel();
        plotJPos = new MultiPlot("Joints Position", "[s]", "[rad] or [m]", "1",
                "2", "3", "4", "5", "6");
        plotJPos.setXRange(0.0, timeDuration);

        plotJVel = new MultiPlot("Joints Velocity", "[s]", "[rad/s]", "1", "2",
                "3", "4", "5", "6");
        plotJVel.setXRange(0.0, timeDuration);

        plotErrOrien = new MultiPlot("End eff orien error", "[s]", "[rad]");
        plotErrOrien.setXRange(0.0, timeDuration);

        plotEndEffErrPos = new MultiPlot("End eff position error", "[s]", "[m]");
        plotEndEffErrPos.setXRange(0.0, timeDuration);

        // robot scene3d panel
        robotPanel = new RobotPanel();
        robotPanel.addRobotToDefaultConfiguration(robot, this);
    }

    @Override
    public void initialize()
    {
        robot.setQ3D(q);
        // end effector path
        trj.interpolate(time, s, ds, dds, quatd, wd);
        // clik
        clik.step(s, ds, quatd, wd);

        q.get(jointData[i]);

        wakeupOn(wakeUp);

        sceneTimer.scheduleAtFixedRate(new Update3DScene(), 0l,
                updateSchedulerTime);

        createPlotPanel();
        updatePlotPanel();

    }

    @Override
    public void processStimulus(Enumeration arg0)
    {
        if(i >= points - 1)
        {
            this.setEnable(false);
            sceneTimer.cancel();
            plotPanel.fillPlots();
            return;
        }

        time += timeStep;
        wakeupOn(wakeUp);

        // end effector path
        trj.interpolate(time, s, ds, dds, quatd, wd);
        // clik
        clik.step(s, ds, quatd, wd);

        i++;
        q.get(jointData[i]);

        updatePlotPanel();
    }

    public final MultiPlotPanel getPlotPanel()
    {
        return plotPanel;
    }

    public final RobotPanel getRobotPanel()
    {
        return robotPanel;
    }

    public final double[][] getJointData()
    {
        return jointData;
    }

    /**
     * Plays the simulation.
     */
    public void play()
    {
    	
    		Timer playTimer = new Timer();
        PlaySimulation playSimulation = new PlaySimulation(robot, jointData, deltaTimer);
        playTimer.scheduleAtFixedRate(playSimulation, 0l, deltaTimer);
    }
    
    /**
     * Starts the simulation.
     */
    public void start()
    {

       setEnable(true);
    }
    
    /**
     * Stops the simulation.
     */
    public void stop()
    {

       setEnable(false);
    }

    private final class Update3DScene extends TimerTask
    {

        @Override
        public void run()
        {
            robot.setQ3DOnly(jointData[i]);

        }

    }

  
    private final void createPlotPanel()
    {

        plotPanel.add(plotJPos);
        plotPanel.add(plotJVel);
        plotPanel.add(plotEndEffErrPos);
        plotPanel.add(plotErrOrien);

        return;
    }

    private final void updatePlotPanel()
    {
        // plotting joints pos
        plotJPos.addVector(time, q, true);
        // plotting joints vel
        plotJVel.addVector(time, qp, true);
        // plot end effector err pos/orien
        plotEndEffErrPos.addPoint(0, time, clik.getPositionError().length(), true);
        plotErrOrien.addPoint(0, time, clik.getOrientationError().length(), true);
    }
}
