/*
 * TipPanel.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.swing;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.robotics.engine.robots.Robot3D;
import javax.robotics.vecmath.RMatrix3d;
import javax.robotics.vecmath.RMatrix4d;
import javax.robotics.vecmath.RVector3d;
import javax.robotics.vecmath.RVector;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * This class creates a panel to set the position and the orientation of robot
 * tip(end effector).
 * 
 * @author <a href="http://www.intermedia.sa.it/lia/">Carmine Lia</a>
 * @since 1.1.0
 * @version 15/10/2005
 */
public class TipPanel extends JPanel implements ChangeListener, ActionListener
{

    /**
     * 
     */
    private static final long serialVersionUID = 3978143257605911607L;

    private double step = 0.01;
    private JSpinner x, y, z, roll, pitch, yaw;
    private JButton update;
    private JLabel errorLabel;
    private Robot3D robot;

    /**
     * Creates the panel of spinners for input forces and torques.
     * 
     * @param robot
     *            the robot to link to.
     * @param stepSize
     *            the minumum size of spinners step.
     */
    public TipPanel(Robot3D robot, double stepSize)
    {
        super();
        this.robot = robot;
        step = stepSize;
        createPanel();
        updateValue();
    }

    private JSpinner createJSpinner(String tip)
    {
        final JSpinner spinner = new JSpinner(new SpinnerNumberModel(0.0,
                -100.0, 100.0, step));
        spinner.addChangeListener(this);
        spinner.setToolTipText(tip);
        // System.out.println(spinner.getPreferredSize());
        spinner.setSize(new Dimension(65, 20));
        return spinner;
    }

    private JPanel createPositionPanel()
    {
        // force panel
        final JPanel fp = new JPanel();
        fp.setLayout(new GridLayout(3, 1, 0, 0));
        /*
         * fp.setBorder(BorderFactory.createTitledBorder(BorderFactory
         * .createLineBorder(Color.gray), "Position"));
         */
        x = createJSpinner("x");
        fp.add(x);

        y = createJSpinner("y");
        fp.add(y);

        z = createJSpinner("z");
        fp.add(z);
        return fp;
    }

    private JPanel createOrientationPanel()
    {
        // torque panel
        final JPanel tp = new JPanel();
        tp.setLayout(new GridLayout(3, 1, 0, 0));
        /*
         * tp.setBorder(BorderFactory.createTitledBorder(BorderFactory
         * .createLineBorder(Color.gray), "Orientation"));
         */
        roll = createJSpinner("roll");
        tp.add(roll);

        pitch = createJSpinner("pitch");
        tp.add(pitch);

        yaw = createJSpinner("yaw");
        tp.add(yaw);
        return tp;
    }

    private void createPanel()
    {
        // create the main panel
        setBorder(BorderFactory.createCompoundBorder(BorderFactory
                .createTitledBorder(BorderFactory.createLineBorder(Color.gray),
                        "Tip"), BorderFactory.createEmptyBorder(2, 5, 2, 5)));
        final BorderLayout layout = new BorderLayout();
        layout.setHgap(2);
        layout.setVgap(5);
        setLayout(layout);
        errorLabel = new JLabel(" ");
        errorLabel.setForeground(Color.RED);
        add(errorLabel, BorderLayout.NORTH);
        add(createPositionPanel(), BorderLayout.WEST);
        add(createOrientationPanel(), BorderLayout.EAST);
        update = new JButton("Set");
      //  update.setSize(100, 30);
        update.addActionListener(this);
        update.setToolTipText("Update the values to the current robot end effector position and orien");
        add(update, BorderLayout.SOUTH);
    }

    /**
     * Updates the panel values to the current robot tip position and
     * orientation.
     */
    public final void updateValue()
    {
        x.removeChangeListener(this);
        y.removeChangeListener(this);
        z.removeChangeListener(this);
        roll.removeChangeListener(this);
        pitch.removeChangeListener(this);
        yaw.removeChangeListener(this);

        final RMatrix3d M = new RMatrix3d();
        final RVector3d p = new RVector3d();
        final RMatrix4d T = new RMatrix4d(robot.kine());
        T.get(M, p);

        x.setValue(p.x);
        y.setValue(p.y);
        z.setValue(p.z);

        roll.setValue(Math.atan2(M.m10, M.m00));
        pitch.setValue(Math.atan2(-M.m20, Math.sqrt(M.m21 * M.m21/*
                                                                     * Math.pow(M.m21,
                                                                     * 2.0)
                                                                     */
                + M.m22 * M.m22 /* Math.pow(M.m22, 2.0) */)));
        yaw.setValue(Math.atan2(M.m21, M.m22));

        x.addChangeListener(this);
        y.addChangeListener(this);
        z.addChangeListener(this);
        roll.addChangeListener(this);
        pitch.addChangeListener(this);
        yaw.addChangeListener(this);
        errorLabel.setText("");
    }

    /**
     * Gets the minimum step of spinners.
     * 
     * @return Returns the step value.
     */
    public final double getStepSize()
    {
        return step;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.swing.event.ChangeListener#stateChanged(javax.swing.event.ChangeEvent)
     */
    public void stateChanged(ChangeEvent e)
    {

        final RMatrix4d T = new RMatrix4d(rpy(), new RVector3d((Double) x
                .getValue(), (Double) y.getValue(), (Double) z.getValue()), 1.0);
        // System.out.println(T);
        final RVector q = robot.invKine(T);//, 0, robot.getDof());
//      System.out.println(q);
        if(q != null)
        {
            errorLabel.setText("");
            if(checkRange(q))
            {
                robot.setQ3D(q);
                errorLabel.setText("");
            }
            else
            {
                errorLabel.setText("IK not in range");
            }
        }
        else
        {
            errorLabel.setText("IK not converge");
        }
    }
    /*
     * (non-Javadoc)
     * 
     * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
     */
    public void actionPerformed(ActionEvent e)
    {
        updateValue();
    }

    /**
     * RPY euler angle to rotation matrix.
     * 
     * @return the rotation matrix.
     */
    private RMatrix3d rpy()
    {
        final RMatrix3d R = new RMatrix3d();
        final double ca, sa, cb, sb, cc, sc;

        ca = Math.cos((Double) roll.getValue());
        sa = Math.sin((Double) roll.getValue());
        cb = Math.cos((Double) pitch.getValue());
        sb = Math.sin((Double) pitch.getValue());
        cc = Math.cos((Double) yaw.getValue());
        sc = Math.sin((Double) yaw.getValue());

        R.m00 = ca * cb;
        R.m01 = ca * sb * sc - sa * cc;
        R.m02 = ca * sb * cc + sa * sc;
        R.m10 = sa * cb;
        R.m11 = sa * sb * sc + ca * cb;
        R.m12 = sa * sb * cc - ca * sc;
        R.m20 = -sb;
        R.m21 = cb * sc;
        R.m22 = cb * cc;

        return R;
    }

    /**
     * Check if the IK respect the joints limits.
     * 
     * @param q
     * @return
     */
    private boolean checkRange(RVector q)
    {
        boolean isInRange = true;
        final int dof = robot.getDof();
        for (int i = 0; i < dof; i++)
            if(q.getElement(i) < robot.getQMin().getElement(i)
                    || q.getElement(i) > robot.getQMax().getElement(i))
            {
                isInRange = false;
                break;
            }
        return isInRange;

    }

}
