/*
 * RMatrix3d.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.vecmath;

import java.io.PrintStream;
import java.util.Locale;

import javax.vecmath.Matrix3d;

/**
 * A double precision floating point 3 by 3 matrix. Primarily to support 3D
 * rotations. This class extends printing functionality of Matrix3d class.
 * 
 * @author <a href="http://www.intermedia.sa.it/lia/">Carmine Lia</a>
 * @since 1.0.1
 * @version 5/10/2005
 */
public class RMatrix3d extends Matrix3d
{

    /**
     * 
     */
    private static final long serialVersionUID = 3257007670102733875L;
    private static final String strLineSeparator = "line.separator";
    private static final String newLine = System.getProperty(strLineSeparator);
    private static final String formatString = "%9.4f ";


    /**
     * Constructs a new matrix with the same values as the Matrix3d parameter.
     * 
     * @param matrix
     *            the 3 x 3 source matrix
     */
    public RMatrix3d(Matrix3d matrix)
    {
        super(matrix);
    }

    /**
     * Constructs a new matrix with the same values as the RMatrix3d parameter.
     * 
     * @param matrix
     *            the 3 x 3 source matrix
     */
    public RMatrix3d(RMatrix3d matrix)
    {
        super(matrix);
    }

    /**
     * Constructs and initializes a RMatrix3d to all zeros.
     */
    public RMatrix3d()
    {
        super();
    }

    /**
     * Constructs and initializes a RMatrix3d from the specified nine element
     * array. this.m00=matrix[0], this.m01=matrix[1], ... , this.m10=matrix[3],
     * etc.
     * 
     * @param matrix
     *            the array of length 9 containing in order.
     */
    public RMatrix3d(double[] matrix)
    {
        super(matrix);
    }
    
    /**
     * Gets S matrix.
     * If R(t) is a rotation matrix the derivate prime R(t)<sup>'</sup>= S(w(t))*R(t), w(t) angular velocity.. 
     * @param w input vector
     */
    public void S(RVector3d w)
    {
        this.m00 =  0;
        this.m01 = -w.z;
        this.m02 =  w.y;
        this.m10 =  w.z;
        this.m11 =  0;
        this.m12 = -w.x;
        this.m20 = -w.y;
        this.m21 =  w.x;
        this.m22 =  0;
    }
    
    /**
     * Get the quaternion from this matrix.
     * @return the quaternion
     */
    public Quaternion getQuaternion()
    {
       Quaternion quaternion = new Quaternion();
       quaternion.w = 0.5*Math.sqrt(this.m00 + this.m11 + this.m22 + 1);
       quaternion.x = 0.5*Math.signum(this.m21-this.m12)*Math.sqrt( this.m00 - this.m11 - this.m22 + 1);
       quaternion.y = 0.5*Math.signum(this.m02-this.m20)*Math.sqrt(-this.m00 + this.m11 - this.m22 + 1);
       quaternion.z = 0.5*Math.signum(this.m10-this.m01)*Math.sqrt(-this.m00 - this.m11 + this.m22 + 1);
       return quaternion;    
    }
    
    /**
     * Prints to standard out
     */
    public void printf()
    {
        this.printf(System.out, formatString);
    }

    /**
     * Prints to standard out according to the format string.
     * <p>
     * The format specifiers for numeric types have the following syntax:
     * 
     * <pre>
     * 
     *  %[width][.precision]conversion
     *  
     * </pre>
     * 
     * <ul>
     * <li> The optional <b><code>width</code></b> is a non-negative decimal
     * integer indicating the minimum number of characters to be written to the
     * output. </li>
     * <li> The optional <b><code>precision</code></b> is a non-negative
     * decimal integer usually used to restrict the number of characters. The
     * specific behavior depends on the conversion. </li>
     * <li> The required <b><code>conversion</code></b> is a character
     * indicating how the argument should be formatted. The set of valid
     * conversions for a given argument depends on the argument's data type.
     * </li>
     * </ul>
     * <p>
     * Example: </br><code>v.printf("%5.3f");</code>
     * </p>
     * 
     * @param format
     *            The format string
     * @see PrintStream#format(java.lang.String, java.lang.Object[])
     */
    public void printf(String format)
    {
        this.printf(System.out, format);
    }

    /**
     * Prints to <code>PrintStream</code>
     * 
     * @param out
     *            The <code>PrintStream</code> object.
     */
    public void printf(PrintStream out)
    {
        this.printf(out, formatString);
    }

    /**
     * Prints to <code>PrintStream</code> according to the format string
     * 
     * @param out
     *            The <code>PrintStream</code> object.
     * @param format
     *            The format string.
     * @see #printf(String)
     */
    public void printf(PrintStream out, String format)
    {
        for (int i = 0; i < 3; i++)
        {
            for (int j = 0; j < 3; j++)
                out.printf(Locale.US, format, this.getElement(i, j));
            out.println(newLine);
        }
    }

    /**
     * Returns a string that contains the values of this RVector.
     * 
     * @return The string.
     */
    public String toString()
    {
        return this.toString(formatString);
    }

    /**
     * Returns a string that contains the values of this RVector, according to
     * the format string
     * 
     * @param format
     *            The format string
     * @return The string
     * @see #printf(String)
     */
    public String toString(String format)
    {

        final StringBuffer buffer = new StringBuffer(72);

        int i, j;

        for (i = 0; i < 3; i++)
        {
            for (j = 0; j < 3; j++)
            {
                buffer.append(String.format(Locale.US, format, this.getElement(i, j)));
                // .append(" ");
            }
            buffer.append(newLine);
        }

        return buffer.toString();
    }
}
