/*
 * RVector4d.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.vecmath;

import java.io.PrintStream;
import java.util.Locale;

import javax.vecmath.Vector4d;

/**
 * A 4-element vector represented by double-precision floating point x,y,z,w
 * coordinates. This class extend printing functionality of Vector4d class.
 * 
 * @author <a href="http://www.intermedia.sa.it/lia/">Carmine Lia</a>
 * @since 1.0.1
 * @version 5/10/2005
 */
public class RVector4d extends Vector4d
{

    /**
     * 
     */
    private static final long serialVersionUID = 3256719580859807287L;

    private static final String strLineSeparator = "line.separator";
    private static final String newLine = System.getProperty(strLineSeparator);
    private static final String formatString = "%9.4f" + newLine;
    
    /**
     * Constructs and initializes a RVector4d to (0,0,0,0).
     */
    public RVector4d()
    {
        this(0.0, 0.0, 0.0, 0.0);
    }
    
    /**
     * Constructs and initializes a RVector4d from the specified RVector4d.
     * 
     * @param vector
     *            the RVector4d containing the initialization x y z w data
     */
    public RVector4d(RVector4d vector)
    {
        this(vector.x, vector.y, vector.z, vector.w);
    }

    /**
     * Constructs and initializes a RVector4d from the specified Vector4d.
     * 
     * @param vector
     *            the Vector4d containing the initialization x y z w data
     */
    public RVector4d(Vector4d vector)
    {
        this(vector.x, vector.y, vector.z, vector.w);
    }

    /**
     * Constructs and initializes a RVector4d from the coordinates contained in
     * the array.
     * 
     * @param vector
     *            the array of length 4 containing x y z w in order
     */
    public RVector4d(double[] vector)
    {
        super(vector);
    }

    /**
     * Constructs and initializes a RVector4d from the specified xyzw
     * coordinates.
     * 
     * @param x
     *            the x coordinate
     * @param y
     *            the y coordinate
     * @param z
     *            the z coordinate
     * @param w
     *            the w coordinate
     */
    public RVector4d(double x, double y, double z, double w)
    {
        super(x, y, z, w);
    }
    

    
    /**
     * Prints to standard out
     */
    public void printf()
    {
        this.printf(System.out, formatString);
    }

    /**
     * Prints to standard out according to the format string.
     * <p>
     * The format specifiers for numeric types have the following syntax:
     * 
     * <pre>
     * %[width][.precision]conversion
     * </pre>
     * 
     * <ul>
     * <li> The optional <b><code>width</code></b> is a non-negative decimal
     * integer indicating the minimum number of characters to be written to the
     * output. </li>
     * <li> The optional <b><code>precision</code></b> is a non-negative
     * decimal integer usually used to restrict the number of characters. The
     * specific behavior depends on the conversion. </li>
     * <li> The required <b><code>conversion</code></b> is a character
     * indicating how the argument should be formatted. The set of valid
     * conversions for a given argument depends on the argument's data type.
     * </li>
     * </ul>
     * <p>
     * Example: </br><code>v.printf("%5.3f");</code>
     * </p>
     * 
     * @param format
     *            The format string
     * @see PrintStream#format(java.lang.String, java.lang.Object[])
     */
    public void printf(String format)
    {
        this.printf(System.out, format);
    }

    /**
     * Prints to <code>PrintStream</code>
     * 
     * @param out
     *            The <code>PrintStream</code> object.
     */
    public void printf(PrintStream out)
    {
        this.printf(out, formatString);
    }

    /**
     * Prints to <code>PrintStream</code> according to the format string
     * 
     * @param out
     *            The <code>PrintStream</code> object.
     * @param format
     *            The format string.
     * @see #printf(String)
     */
    public void printf(PrintStream out, String format)
    {
        out.printf(Locale.US, "[x] " + format, this.x);
        out.printf(Locale.US, "[y] " + format, this.y);
        out.printf(Locale.US, "[z] " + format, this.z);
        out.printf(Locale.US, "[w] " + format, this.w);

    }

    /**
     * Returns a string that contains the values of this RVector.
     * 
     * @return The string.
     */
    public String toString()
    {
        return this.toString(formatString);
    }

    /**
     * Returns a string that contains the values of this RVector, according to
     * the format string
     * 
     * @param format
     *            The format string
     * @return The string
     * @see #printf(String)
     */
    public String toString(String format)
    {
        final StringBuffer buffer = new StringBuffer(32);

        buffer.append(String.format(Locale.US, "[x] " + format, this.x));
        buffer.append(String.format(Locale.US, "[y] " + format, this.y));
        buffer.append(String.format(Locale.US, "[z] " + format, this.z));
        buffer.append(String.format(Locale.US, "[w] " + format, this.w));

        return buffer.toString();

    }
}
