package javax.robotics.vecmath.jama;

import javax.robotics.vecmath.jama.util.*;

/**
 * QR Decomposition.
 * <P>
 * For an m-by-n matrix A with m >= n, the QR decomposition is an m-by-n
 * orthogonal matrix Q and an n-by-n upper triangular matrix R so that A = Q*R.
 * <P>
 * The QR decompostion always exists, even if the matrix does not have full
 * rank, so the constructor will never fail. The primary use of the QR
 * decomposition is in the least squares solution of nonsquare systems of
 * simultaneous linear equations. This will fail if isFullRank() returns false.
 */

public class QRDecomposition implements java.io.Serializable
{

    /*
     * ------------------------ Class variables ------------------------
     */

    /**
     * 
     */
    private static final long serialVersionUID = 4121138013284741682L;

    /**
     * Array for internal storage of decomposition.
     * 
     * @serial internal array storage.
     */
    private double[][] QR;

    /**
     * Row and column dimensions.
     * 
     * @serial column dimension.
     * @serial row dimension.
     */
    private int m, n;

    /**
     * Array for internal storage of diagonal of R.
     * 
     * @serial diagonal of R.
     */
    private double[] Rdiag;

    /*
     * ------------------------ Constructor ------------------------
     */

    /**
     * QR Decomposition, computed by Householder reflections.
     * 
     * @param A
     *            Rectangular matrix
     */

    public QRDecomposition(double[][] A, int NumRow, int NumCol)
    {
        // Initialize.
        QR = A;
        m = NumRow;
        n = NumCol;
        Rdiag = new double[n];

        // Main loop.
        for (int k = 0; k < n; k++)
        {
            // Compute 2-norm of k-th column without under/overflow.
            double nrm = 0;
            for (int i = k; i < m; i++)
            {
                nrm = Maths.hypot(nrm, QR[i][k]);
            }

            if(nrm != 0.0)
            {
                // Form k-th Householder vector.
                if(QR[k][k] < 0)
                {
                    nrm = -nrm;
                }
                for (int i = k; i < m; i++)
                {
                    QR[i][k] /= nrm;
                }
                QR[k][k] += 1.0;

                // Apply transformation to remaining columns.
                for (int j = k + 1; j < n; j++)
                {
                    double s = 0.0;
                    for (int i = k; i < m; i++)
                    {
                        s += QR[i][k] * QR[i][j];
                    }
                    s = -s / QR[k][k];
                    for (int i = k; i < m; i++)
                    {
                        QR[i][j] += s * QR[i][k];
                    }
                }
            }
            Rdiag[k] = -nrm;
        }
    }

    /*
     * ------------------------ Public Methods ------------------------
     */

    /**
     * Is the matrix full rank?
     * 
     * @return true if R, and hence A, has full rank.
     */

    public boolean isFullRank()
    {
        for (int j = 0; j < n; j++)
        {
            if(Rdiag[j] == 0)
                return false;
        }
        return true;
    }

    /**
     * Return the Householder vectors
     * 
     * @return Lower trapezoidal matrix whose columns define the reflections
     */

    public double[][] getH()
    {
        double[][] H = new double[m][n];
        for (int i = 0; i < m; i++)
        {
            for (int j = 0; j < n; j++)
            {
                if(i >= j)
                {
                    H[i][j] = QR[i][j];
                }
                else
                {
                    H[i][j] = 0.0;
                }
            }
        }
        return H;
    }

    /**
     * Return the upper triangular factor
     * 
     * @return R
     */

    public double[][] getR()
    {
        double[][] R = new double[n][n];
        for (int i = 0; i < n; i++)
        {
            for (int j = 0; j < n; j++)
            {
                if(i < j)
                {
                    R[i][j] = QR[i][j];
                }
                else
                    if(i == j)
                    {
                        R[i][j] = Rdiag[i];
                    }
                    else
                    {
                        R[i][j] = 0.0;
                    }
            }
        }
        return R;
    }

    /**
     * Generate and return the (economy-sized) orthogonal factor
     * 
     * @return Q
     */

    public double[][] getQ()
    {
        double[][] Q = new double[m][n];
        for (int k = n - 1; k >= 0; k--)
        {
            for (int i = 0; i < m; i++)
            {
                Q[i][k] = 0.0;
            }
            Q[k][k] = 1.0;
            for (int j = k; j < n; j++)
            {
                if(QR[k][k] != 0)
                {
                    double s = 0.0;
                    for (int i = k; i < m; i++)
                    {
                        s += QR[i][k] * Q[i][j];
                    }
                    s = -s / QR[k][k];
                    for (int i = k; i < m; i++)
                    {
                        Q[i][j] += s * QR[i][k];
                    }
                }
            }
        }
        return Q;
    }

    /**
     * Least squares solution of A*X = B
     * 
     * @param B
     *            A matrix with as many rows as A and any number of columns.
     * @return X that minimizes the two norm of Q*R*X-B.
     * @exception IllegalArgumentException
     *                Matrix row dimensions must agree.
     * @exception RuntimeException
     *                Matrix is rank deficient.
     */

    public double[][] solve(double[][] B, int NumRow, int NumCol)
    {
        if(NumRow != m)
        {
            throw new IllegalArgumentException(
                    "Matrix row dimensions must agree.");
        }
        if(!this.isFullRank())
        {
            throw new RuntimeException("Matrix is rank deficient.");
        }

        // Copy right hand side
        int nx = NumCol;
        double[][] X = B;

        // Compute Y = transpose(Q)*B
        for (int k = 0; k < n; k++)
        {
            for (int j = 0; j < nx; j++)
            {
                double s = 0.0;
                for (int i = k; i < m; i++)
                {
                    s += QR[i][k] * X[i][j];
                }
                s = -s / QR[k][k];
                for (int i = k; i < m; i++)
                {
                    X[i][j] += s * QR[i][k];
                }
            }
        }
        // Solve R*X = Y;
        for (int k = n - 1; k >= 0; k--)
        {
            for (int j = 0; j < nx; j++)
            {
                X[k][j] /= Rdiag[k];
            }
            for (int i = 0; i < k; i++)
            {
                for (int j = 0; j < nx; j++)
                {
                    X[i][j] -= X[k][j] * QR[i][k];
                }
            }
        }
        return getMatrix(X, 0, n - 1, 0, nx - 1);
    }
    
    /** Get a submatrix.
    @param i0   Initial row index
    @param i1   Final row index
    @param j0   Initial column index
    @param j1   Final column index
    @return     A(i0:i1,j0:j1)
    @exception  ArrayIndexOutOfBoundsException Submatrix indices
    */

    private double[][] getMatrix (double[][] A, int i0, int i1, int j0, int j1) {
       
       double[][] B = new double[i1-i0+1][j1-j0+1];
       try {
          for (int i = i0; i <= i1; i++) {
             for (int j = j0; j <= j1; j++) {
                B[i-i0][j-j0] = A[i][j];
             }
          }
       } catch(ArrayIndexOutOfBoundsException e) {
          throw new ArrayIndexOutOfBoundsException("Submatrix indices");
       }
       return B;
    }
}
