/*
 * XmlRobot.java -
 * Copyright (c) 2005 Carmine Lia
 * e-mail:      carmine.lia@libero.it
 * web-address: http://www.intermedia.sa.it/lia

 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package javax.robotics.xml;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;

import org.apache.xmlbeans.XmlException;

import robotML.DynamicsType;
import robotML.LinkType;
import robotML.MotorType;
import robotML.RobotConfigurationDocument;
/**
 * XML handler class for robot configuration file.
 * 
 * @author <a href="http://www.intermedia.sa.it/lia/">Carmine Lia</a>
 * @since 1.1.0
 * @version 30/10/2005
 */
public class XmlRobot
{
    private static final int INITROBOT_DIM = 19;
    private static final int INITMOTOR_DIM = 4;

    private RobotConfigurationDocument robotDoc = null;
    private File robotXml;
    private InputStream robotXmlStream;
    private String description, name;
    private int dof, nLinks;
    private boolean motor, dynamics;
    private double[] linkData, motorData, qMin, qMax, qOffset, a, d, teta, alfa;
    private boolean[] type, immobile;

    /**
     * Constructor of the xml handler.
     * 
     * @param robotXml
     *            the path of xml configuration file.
     * @throws XmlException
     * @throws IOException
     */
    public XmlRobot(String robotXml) throws XmlException, IOException
    {
        this(new File(robotXml));
    }
    
    /**
     * Constructor of the xml handler from file.
     * 
     * @param robotXml
     *            the path of xml configuration file.
     * @throws XmlException
     * @throws IOException
     */
    public XmlRobot(File robotXml) throws XmlException, IOException
    {
        this.robotXml = robotXml;
        robotDoc = RobotConfigurationDocument.Factory.parse(this.robotXml);
        if(!robotDoc.validate())
            throw new XmlException("The file " + robotXml.getName()
                    + " isn't a valid xml");

        readGeneralData();
        readLinkData();
    }
    /**
     * Constructor of the xml handler from input stream.
     * 
     * @param robotXml
     *            the path of xml configuration file.
     * @throws XmlException
     * @throws IOException
     */
    public XmlRobot(InputStream robotXml) throws XmlException, IOException
    {
        this.robotXmlStream = robotXml;
        robotDoc = RobotConfigurationDocument.Factory.parse(this.robotXmlStream);
        if(!robotDoc.validate())
            throw new XmlException("The input stream isn't a valid xml");

        readGeneralData();
        readLinkData();
    }
    /**
     * read the information about robot.
     */
    private void readGeneralData()
    {
        description = robotDoc.getRobotConfiguration().getDescription();
        name = robotDoc.getRobotConfiguration().getName();
        dof = robotDoc.getRobotConfiguration().getDof();
        dynamics = robotDoc.getRobotConfiguration().getDynamics();
        motor = robotDoc.getRobotConfiguration().getMotor();
    }

    /**
     * read the link data and the dynamics data
     */
    private void readLinkData()
    {
        LinkType[] link;
        MotorType linkMotor;
        DynamicsType linkDyna;

        link = robotDoc.getRobotConfiguration().getLinkArray();
        nLinks = link.length;
        qMin = new double[nLinks];
        qMax = new double[nLinks];
        qOffset = new double[nLinks];
        linkData = new double[nLinks * INITROBOT_DIM];
        motorData = new double[nLinks * INITMOTOR_DIM];
        a = new double[nLinks];
        d = new double[nLinks];
        alfa = new double[nLinks];
        teta = new double[nLinks];
        type = new boolean[nLinks];
        immobile = new boolean[nLinks];
        
        int j = 0;
        int k = 0;
        int type;
        for (int i = 0; i < link.length; i++)
        {
            linkData[j++] = type = (link[i].getJointType()
                    .equalsIgnoreCase("revolute")) ? 0 : 1;

            linkData[j++] = teta[i] = link[i].getTheta();
            linkData[j++] = d[i] = link[i].getD();
            linkData[j++] = a[i] = link[i].getA();
            linkData[j++] = alfa[i] = link[i].getAlpha();
            if(type == 0)
            {
                this.type[i] = false;
                linkData[j++] = link[i].getThetaMin();
                qMin[i] = link[i].getThetaMin();
                linkData[j++] = link[i].getThetaMax();
                qMax[i] = link[i].getThetaMax();
            }
            else
            {
                this.type[i] = true;
                linkData[j++] = link[i].getDMin();
                qMin[i] = link[i].getDMin();
                linkData[j++] = link[i].getDMax();
                qMax[i] = link[i].getDMax();
            }
            linkData[j++] = link[i].getJointOffset();
            qOffset[i] = link[i].getJointOffset();
            if(dynamics)
            {
                linkDyna = link[i].getDynamics();

                linkData[j++] = linkDyna.getM();
                linkData[j++] = linkDyna.getCx();
                linkData[j++] = linkDyna.getCy();
                linkData[j++] = linkDyna.getCz();
                linkData[j++] = linkDyna.getIxx();
                linkData[j++] = linkDyna.getIxy();
                linkData[j++] = linkDyna.getIxz();
                linkData[j++] = linkDyna.getIyy();
                linkData[j++] = linkDyna.getIyz();
                linkData[j++] = linkDyna.getIzz();
                if((motor) && (dynamics))
                {
                    linkMotor = linkDyna.getMotor();
                    motorData[k++] = linkMotor.getIm();
                    motorData[k++] = linkMotor.getGr();
                    motorData[k++] = linkMotor.getB();
                    motorData[k++] = linkMotor.getCf();
                }
            }
            else
            {
                linkData[j++] = 0;
                linkData[j++] = 0;
                linkData[j++] = 0;
                linkData[j++] = 0;
                linkData[j++] = 0;
                linkData[j++] = 0;
                linkData[j++] = 0;
                linkData[j++] = 0;
                linkData[j++] = 0;
                linkData[j++] = 0;
            }
            linkData[j++] = (link[i].getImmobile()) ? 1 : 0;
            immobile[i] = (link[i].getImmobile()) ? true : false;

        }
    }

    /**
     * Gets the tag description of the file.
     * 
     * @return the tag description value.
     */
    final public String getDescription()
    {
        return description;
    }

    /**
     * Gets the tag dof of the file.
     * 
     * @return the value of the tag dof.
     */
    final public int getDof()
    {
        return dof;
    }

    /**
     * Gets the array of theta-min or d-min tags of the links.
     * 
     * @return the value of the tags theta-min or d-min.
     */
    final public double[] getQMin()
    {
        return qMin;
    }

    /**
     * Gets the array of theta-max or d-max tags of the links.
     * 
     * @return the value of the tags theta-max or d-max.
     */
    final public double[] getQMax()
    {
        return qMax;
    }
    /**
     * Gets the array of offset tag of the links.
     * 
     * @return the value of the tag offset.
     */
    final public double[] getQOffset()
    {
        return qOffset;
    }
    /**
     * Gets the array of data about all the links of the robot.
     * 
     * @return the array nLinks-by-19 of the links data.
     */
    final public double[] getLinkData()
    {
        return linkData;
    }

    /**
     * Gets the tag motor of file.
     * 
     * @return true if file contains the motor data.
     */
    final public boolean isMotorSet()
    {
        return motor;
    }

    /**
     * Gets the tag dynamics of file.
     * 
     * @return true if file contains dynamics data.
     */
    final public boolean isDynamicsSet()
    {
        return dynamics;
    }

    /**
     * Gets the motor data of the links.
     * 
     * @return the nLinks-by-4 array of motor data.
     */
    final public double[] getMotorData()
    {
        return motorData;
    }

    /**
     * Gets the tag name of the file.
     * 
     * @return the value of the tag name.
     */
    final public String getName()
    {
        return name;
    }

    /**
     * Gets the number of links
     * 
     * @return the number of links.
     */
    final public int getNLinks()
    {
        return nLinks;
    }
    
    /**
     * Gets the DH parameters a.
     * 
     * @return the array of parameters a.
     */
    public double[] getA()
    {
        return a;
    }
    /**
     * Gets the DH parameters &alpha;.
     * 
     * @return the array parameters &alpha;.
     */
    public double[] getAlfa()
    {
        return alfa;
    }
    /**
     * Gets the DH parameters d.
     * 
     * @return the array of parameters d.
     */
    public double[] getD()
    {
        return d;
    }
    /**
     * Gets the immobile flags.
     * 
     * @return the array of immobile flags.
     */
    public boolean[] getImmobile()
    {
        return immobile;
    }
    /**
     * Gets the DH parameters &theta;.
     * 
     * @return the array parameters &theta;.
     */
   
    public double[] getTeta()
    {
        return teta;
    }
    /**
     * Gets the joint type flags.
     * true if prismatic.
     * @return the array of joint type flags.
     */
    public boolean[] getType()
    {
        return type;
    }

    /**
     * Gets the handle of xml configuration file.
     * 
     * @return the xml handler.
     */
    final public RobotConfigurationDocument getXmlHandler()
    {
        return robotDoc;
    }

}
